@@ -1,5 +1,5 @@
 /* darkstat 3
- * copyright (c) 2001-2007 Emil Mikulic.
+ * copyright (c) 2001-2008 Emil Mikulic.
  *
  * err.h: BSD-like err() and warn() functions
  *
@@ -16,12 +16,13 @@
  * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
  */
 
-void err(const int code, const char *format, ...);
-void errx(const int code, const char *format, ...);
+void err(const int code, const char *format, ...) _noreturn_;
+void errx(const int code, const char *format, ...) _noreturn_;
 
 void warn(const char *format, ...);
 void warnx(const char *format, ...);
 
+extern int want_verbose;
 void verbosef(const char *format, ...);
 void dverbosef(const char *format _unused_, ...);
 
