@@ -16,6 +16,7 @@
 #include "db.h"
 #include "html.h"
 #include "ncache.h"
+#include "now.h"
 
 #include <arpa/inet.h> /* inet_aton() */
 #include <assert.h>
@@ -37,7 +38,8 @@ typedef const void * (key_func_t)(const struct bucket *);
 typedef int (find_func_t)(const struct bucket *, const void *);
 typedef struct bucket * (make_func_t)(const void *);
 typedef void (format_cols_func_t)(struct str *);
-typedef void (format_row_func_t)(struct str *, const struct bucket *);
+typedef void (format_row_func_t)(struct str *, const struct bucket *,
+   const char *);
 
 struct hashtable {
    uint8_t bits;     /* size of hashtable in bits */
@@ -201,6 +203,7 @@ make_func_host(const void *key)
    MAKE_BUCKET(b, h, host);
    h->ip = CASTKEY(in_addr_t);
    h->dns = NULL;
+   h->last_seen = now;
    memset(&h->mac_addr, 0, sizeof(h->mac_addr));
    h->ports_tcp = NULL;
    h->ports_udp = NULL;
@@ -261,56 +264,67 @@ format_cols_host(struct str *buf)
     * when setting sort direction
     */
    str_append(buf,
-      "<table border=\"1\">\n"
+      "<table>\n"
       "<tr>\n"
       " <th>IP</th>\n"
       " <th>Hostname</th>\n");
-   if (show_mac_addrs)
-   str_append(buf,
+   if (show_mac_addrs) str_append(buf,
       " <th>MAC Address</th>\n");
    str_append(buf,
       " <th><a href=\"?sort=in\">In</a></th>\n"
       " <th><a href=\"?sort=out\">Out</a></th>\n"
       " <th><a href=\"?sort=total\">Total</a></th>\n"
+      " <th>Last seen</th>\n"
       "</tr>\n");
 }
 
 static void
-format_row_host(struct str *buf, const struct bucket *b)
+format_row_host(struct str *buf, const struct bucket *b,
+   const char *css_class)
 {
    const char *ip = ip_to_str( b->u.host.ip );
+   struct str *lastseen = NULL;
+   time_t last_t;
+
+   str_appendf(buf,
+      "<tr class=\"%s\">\n"
+      " <td><a href=\"%s/\">%s</a></td>\n"
+      " <td>%s</td>\n",
+      css_class,
+      ip, ip,
+      (b->u.host.dns == NULL) ? "" : b->u.host.dns);
 
    if (show_mac_addrs)
+      str_appendf(buf,
+         " <td><tt>%x:%x:%x:%x:%x:%x</tt></td>\n",
+         b->u.host.mac_addr[0],
+         b->u.host.mac_addr[1],
+         b->u.host.mac_addr[2],
+         b->u.host.mac_addr[3],
+         b->u.host.mac_addr[4],
+         b->u.host.mac_addr[5]);
+
+   last_t = b->u.host.last_seen;
+   if (now >= last_t)
+      lastseen = length_of_time(now - last_t);
+
    str_appendf(buf,
-      "<tr>\n"
-      " <td><a href=\"%s/\">%s</a></td>\n"
-      " <td>%s</td>\n"
-      " <td><tt>%x:%x:%x:%x:%x:%x</tt></td>\n"
-      " <td>%'qu</td>\n"
-      " <td>%'qu</td>\n"
-      " <td>%'qu</td>\n"
-      "</tr>\n",
-      ip, ip, (b->u.host.dns == NULL)?"":b->u.host.dns,
-      b->u.host.mac_addr[0],
-      b->u.host.mac_addr[1],
-      b->u.host.mac_addr[2],
-      b->u.host.mac_addr[3],
-      b->u.host.mac_addr[4],
-      b->u.host.mac_addr[5],
-      b->in, b->out, b->total
-   );
-   else
+      " <td class=\"num\">%'qu</td>\n"
+      " <td class=\"num\">%'qu</td>\n"
+      " <td class=\"num\">%'qu</td>\n"
+      " <td class=\"num\">",
+      b->in, b->out, b->total);
+
+   if (lastseen == NULL)
+      str_append(buf, "(clock error)");
+   else {
+      str_appendstr(buf, lastseen);
+      str_free(lastseen);
+   }
+
    str_appendf(buf,
-      "<tr>\n"
-      " <td><a href=\"%s/\">%s</a></td>\n"
-      " <td>%s</td>\n"
-      " <td>%'qu</td>\n"
-      " <td>%'qu</td>\n"
-      " <td>%'qu</td>\n"
-      "</tr>\n",
-      ip, ip, (b->u.host.dns == NULL)?"":b->u.host.dns,
-      b->in, b->out, b->total
-   );
+       "</td>\n"
+      "</tr>\n");
 
    /* Only resolve hosts "on demand" */
    if (b->u.host.dns == NULL)
@@ -321,32 +335,34 @@ static void
 format_cols_port_tcp(struct str *buf)
 {
    str_append(buf,
-      "<table border=\"1\">\n"
+      "<table>\n"
       "<tr>\n"
-      " <td>Port</td>\n"
-      " <td>Service</td>\n"
-      " <td>In</td>\n"
-      " <td>Out</td>\n"
-      " <td>Total</td>\n"
-      " <td>SYNs</td>\n"
+      " <th>Port</td>\n"
+      " <th>Service</td>\n"
+      " <th>In</td>\n"
+      " <th>Out</td>\n"
+      " <th>Total</td>\n"
+      " <th>SYNs</td>\n"
       "</tr>\n"
    );
 }
 
 static void
-format_row_port_tcp(struct str *buf, const struct bucket *b)
+format_row_port_tcp(struct str *buf, const struct bucket *b,
+   const char *css_class)
 {
    const struct port_tcp *p = &(b->u.port_tcp);
 
    str_appendf(buf,
-      "<tr>\n"
-      " <td>%u</td>\n"
+      "<tr class=\"%s\">\n"
+      " <td class=\"num\">%u</td>\n"
       " <td>%s</td>\n"
-      " <td>%'qu</td>\n"
-      " <td>%'qu</td>\n"
-      " <td>%'qu</td>\n"
-      " <td>%'qu</td>\n"
+      " <td class=\"num\">%'qu</td>\n"
+      " <td class=\"num\">%'qu</td>\n"
+      " <td class=\"num\">%'qu</td>\n"
+      " <td class=\"num\">%'qu</td>\n"
       "</tr>\n",
+      css_class,
       p->port, getservtcp(p->port), b->in, b->out, b->total, p->syn
    );
 }
@@ -355,30 +371,32 @@ static void
 format_cols_port_udp(struct str *buf)
 {
    str_append(buf,
-      "<table border=\"1\">\n"
+      "<table>\n"
       "<tr>\n"
-      " <td>Port</td>\n"
-      " <td>Service</td>\n"
-      " <td>In</td>\n"
-      " <td>Out</td>\n"
-      " <td>Total</td>\n"
+      " <th>Port</td>\n"
+      " <th>Service</td>\n"
+      " <th>In</td>\n"
+      " <th>Out</td>\n"
+      " <th>Total</td>\n"
       "</tr>\n"
    );
 }
 
 static void
-format_row_port_udp(struct str *buf, const struct bucket *b)
+format_row_port_udp(struct str *buf, const struct bucket *b,
+   const char *css_class)
 {
    const struct port_udp *p = &(b->u.port_udp);
 
    str_appendf(buf,
-      "<tr>\n"
-      " <td>%u</td>\n"
+      "<tr class=\"%s\">\n"
+      " <td class=\"num\">%u</td>\n"
       " <td>%s</td>\n"
-      " <td>%'qu</td>\n"
-      " <td>%'qu</td>\n"
-      " <td>%'qu</td>\n"
+      " <td class=\"num\">%'qu</td>\n"
+      " <td class=\"num\">%'qu</td>\n"
+      " <td class=\"num\">%'qu</td>\n"
       "</tr>\n",
+      css_class,
       p->port, getservudp(p->port), b->in, b->out, b->total
    );
 }
@@ -387,30 +405,32 @@ static void
 format_cols_ip_proto(struct str *buf)
 {
    str_append(buf,
-      "<table border=\"1\">\n"
+      "<table>\n"
       "<tr>\n"
-      " <td>#</td>\n"
-      " <td>Protocol</td>\n"
-      " <td>In</td>\n"
-      " <td>Out</td>\n"
-      " <td>Total</td>\n"
+      " <th>#</td>\n"
+      " <th>Protocol</td>\n"
+      " <th>In</td>\n"
+      " <th>Out</td>\n"
+      " <th>Total</td>\n"
       "</tr>\n"
    );
 }
 
 static void
-format_row_ip_proto(struct str *buf, const struct bucket *b)
+format_row_ip_proto(struct str *buf, const struct bucket *b,
+   const char *css_class)
 {
    const struct ip_proto *p = &(b->u.ip_proto);
 
    str_appendf(buf,
-      "<tr>\n"
-      " <td>%u</td>\n"
+      "<tr class=\"%s\">\n"
+      " <td class=\"num\">%u</td>\n"
       " <td>%s</td>\n"
-      " <td>%'qu</td>\n"
-      " <td>%'qu</td>\n"
-      " <td>%'qu</td>\n"
+      " <td class=\"num\">%'qu</td>\n"
+      " <td class=\"num\">%'qu</td>\n"
+      " <td class=\"num\">%'qu</td>\n"
       "</tr>\n",
+      css_class,
       p->proto, getproto(p->proto),
       b->in, b->out, b->total
    );
@@ -497,8 +517,6 @@ hashtable_rehash(struct hashtable *h, const uint8_t bits)
 
    free(h->table);
    h->table = new_table;
-   verbosef("hashtable rehashed to %d slots (was using %d of %d)",
-      h->size, h->count, old_size);
 }
 
 static void
@@ -552,9 +570,9 @@ hashtable_find_or_insert(struct hashtable *h, const void *key)
    struct bucket *b = hashtable_search(h, key);
 
    if (b == NULL) {
-      /* Not found so insert.  But first check occupancy. */
-      assert(h->count <= h->count_max);
-      if (h->count == h->count_max) hashtable_reduce(h);
+      /* Not found, so insert after checking occupancy. */
+      /*assert(h->count <= h->count_max);*/
+      if (h->count >= h->count_max) hashtable_reduce(h);
       b = h->make_func(key);
       hashtable_insert(h, b);
    }
@@ -764,10 +782,6 @@ host_get_ip_proto(struct bucket *host, const uint8_t proto)
 
 static struct str *html_hosts_main(const char *qs);
 static struct str *html_hosts_detail(const char *ip);
-static struct str *html_hashtable_stats(struct hashtable *h);
-static struct str *html_tcp_stats(const char *ip);
-static struct str *html_udp_stats(const char *ip);
-static struct str *html_proto_stats(const char *ip);
 
 /* ---------------------------------------------------------------------------
  * Web interface: delegate the /hosts/ space.
@@ -785,21 +799,9 @@ html_hosts(const char *uri, const char *query)
    if (num_elems == 1)
       /* /hosts/ */
       buf = html_hosts_main(query);
-   else if ((num_elems == 2) && (strcmp(elem[1], "stats") == 0))
-      /* /hosts/stats/ */
-      buf = html_hashtable_stats(hosts_db);
    else if (num_elems == 2)
       /* /hosts/<IP of host>/ */
       buf = html_hosts_detail(elem[1]);
-   else if ((num_elems == 3) && (strcmp(elem[2], "tcpstats") == 0))
-      /* /hosts/<IP of host>/tcpstats/ */
-      buf = html_tcp_stats(elem[1]);
-   else if ((num_elems == 3) && (strcmp(elem[2], "udpstats") == 0))
-      /* /hosts/<IP of host>/udpstats/ */
-      buf = html_udp_stats(elem[1]);
-   else if ((num_elems == 3) && (strcmp(elem[2], "protostats") == 0))
-      /* /hosts/<IP of host>/protostats/ */
-      buf = html_proto_stats(elem[1]);
 
    for (i=0; i<num_elems; i++)
       free(elem[i]);
@@ -817,6 +819,7 @@ format_table(struct str *buf, struct hashtable *ht, int start,
 {
    const struct bucket **table;
    uint32_t i, pos, end;
+   int alt = 0;
 
    if ((ht == NULL) || (ht->count == 0)) {
       str_append(buf, "<p>The table is empty.</p>\n");
@@ -841,11 +844,14 @@ format_table(struct str *buf, struct hashtable *ht, int start,
    } else
       end = min(ht->count, (uint32_t)start+MAX_ENTRIES);
 
-   str_appendf(buf, "(%u-%u of %u)<br/>", start+1, end, ht->count);
+   str_appendf(buf, "(%u-%u of %u)<br/>\n", start+1, end, ht->count);
    qsort_buckets(table, ht->count, start, end, sort);
    ht->format_cols_func(buf);
-   for (i=start; i<end; i++)
-      ht->format_row_func(buf, table[i]);
+
+   for (i=start; i<end; i++) {
+      ht->format_row_func(buf, table[i], alt ? "alt1" : "alt2");
+      alt = !alt; /* alternate class for table rows */
+   }
    free(table);
    str_append(buf, "</table>\n");
 }
@@ -902,7 +908,7 @@ html_hosts_main(const char *qs)
 #define FULL "full table"
 
    str_append(buf, html_header_1);
-   str_appendf(buf, "<title>darkstat3: Hosts (%s)</title>\n", interface);
+   str_appendf(buf, " <title>darkstat3: Hosts (%s)</title>\n", interface);
    str_append(buf, html_header_2);
    str_appendf(buf, "<h2 class=\"pageheader\">Hosts (%s)</h2>\n", interface);
    format_table(buf, hosts_db, start, sort, full);
@@ -923,13 +929,11 @@ html_hosts_main(const char *qs)
       str_appendf(buf, " | <a href=\"?full=yes&sort=%s\">" FULL "</a>",
          sortstr);
 
-   str_append(buf, " | <a href=\"stats/\">statistics</a> | ");
-
    if (start+MAX_ENTRIES < (int)hosts_db->count)
-      str_appendf(buf, "<a href=\"?start=%d&sort=%s\">" NEXT "</a>",
+      str_appendf(buf, " | <a href=\"?start=%d&sort=%s\">" NEXT "</a>",
          start+MAX_ENTRIES, sortstr);
    else
-      str_append(buf, NEXT);
+      str_append(buf, " | " NEXT);
 
    str_append(buf, "<br/>\n");
    str_append(buf, html_footer);
@@ -949,7 +953,9 @@ static struct str *
 html_hosts_detail(const char *ip)
 {
    struct bucket *h;
-   struct str *buf;
+   struct str *buf, *ls_len;
+   char ls_when[100];
+   time_t ls;
 
    h = host_search(ip);
    if (h == NULL)
@@ -967,6 +973,10 @@ html_hosts_detail(const char *ip)
       ip,
       (h->u.host.dns == NULL)?"(resolving...)":h->u.host.dns);
 
+   /* Resolve host "on demand" */
+   if (h->u.host.dns == NULL)
+      dns_queue(h->u.host.ip);
+
    if (show_mac_addrs)
    str_appendf(buf,
       "<b>MAC Address:</b> "
@@ -978,6 +988,26 @@ html_hosts_detail(const char *ip)
       h->u.host.mac_addr[4],
       h->u.host.mac_addr[5]);
 
+   str_append(buf,
+      "</p>\n"
+      "<p>\n"
+      "<b>Last seen:</b> ");
+
+   ls = h->u.host.last_seen;
+   if (strftime(ls_when, sizeof(ls_when),
+      "%Y-%m-%d %H:%M:%S %Z%z", localtime(&ls)) != 0)
+         str_append(buf, ls_when);
+
+   if (h->u.host.last_seen <= now) {
+      ls_len = length_of_time(now - h->u.host.last_seen);
+      str_append(buf, " (");
+      str_appendstr(buf, ls_len);
+      str_free(ls_len);
+      str_append(buf, " ago)");
+   } else {
+      str_append(buf, " (in the future, possible clock problem)");
+   }
+
    str_appendf(buf,
       "</p>\n"
       "<p>\n"
@@ -989,135 +1019,34 @@ html_hosts_detail(const char *ip)
 
    str_append(buf, "<h3>TCP ports</h3>\n");
    format_table(buf, h->u.host.ports_tcp, 0,TOTAL,0);
-   str_append(buf, "<a href=\"tcpstats/\">statistics</a><br/>\n");
 
    str_append(buf, "<h3>UDP ports</h3>\n");
    format_table(buf, h->u.host.ports_udp, 0,TOTAL,0);
-   str_append(buf, "<a href=\"udpstats/\">statistics</a><br/>\n");
 
    str_append(buf, "<h3>IP protocols</h3>\n");
    format_table(buf, h->u.host.ip_protos, 0,TOTAL,0);
-   str_append(buf, "<a href=\"protostats/\">statistics</a><br/>\n");
 
    str_append(buf, html_footer);
    return (buf);
 }
 
-/* ---------------------------------------------------------------------------
- * Web interface: statistics of a given hashtable.
- */
-static struct str *
-html_hashtable_stats(struct hashtable *h)
-{
-   struct str *buf;
-   uint32_t i, max_len = 0, num_empty = 0, ovr_buckets = 0, ovr_slots = 0;
-
-   buf = str_make();
-   str_appendf(buf,
-      "%s"
-      " <title>Hashtable histogram</title>\n"
-      " <style type=\"text/css\">\n"
-      "  thead td { font-weight:bold; background:#E0E0E0; }\n"
-      "  tbody td { text-align:right; background:#F0F0F0; }\n"
-      " </style>\n"
-      "%s"
-      "<h2>Hashtable statistics</h2>\n",
-      html_header_1, html_header_2
-   );
-
-   if (h == NULL)
-      str_append(buf, "<p>Table is empty.</p>\n");
-   else {
-      /* Traverse table. */
-      for (i=0; i<h->size; i++)
-      {
-         struct bucket *b = h->table[i];
-         uint32_t curr_len = 0;
-
-         if (b == NULL)
-            num_empty++;
-         else
-            for (b = h->table[i]; b != NULL; b = b->next)
-               curr_len++;
-
-         if (curr_len > 1)
-         {
-            ovr_slots++;
-            ovr_buckets += curr_len - 1;
-            max_len = max(max_len, curr_len);
-         }
-      }
-
-      str_appendf(buf,
-         "<p>\n"
-         "<b>Slots:</b> %'u<br/>\n"
-         "<b>Records:</b> %'u<br/>\n"
-         "<b>Slots empty:</b> %'u<br/>\n"
-         "<b>Slots overflowed:</b> %'u<br/>\n"
-         "<b>Overflow buckets:</b> %'u<br/>\n"
-         "<b>Longest chain:</b> %'u<br/>\n"
-         "<b>Usage:</b> %u%%<br/>\n"
-         "<br/>\n"
-         "<b>Inserts:</b> %'qu<br/>\n"
-         "<b>Searches:</b> %'qu<br/>\n"
-         "<b>Deletions:</b> %'qu<br/>\n"
-         "<b>Rehashes:</b> %'qu<br/>\n"
-         "</p>\n"
-         ,
-         h->size, h->count, num_empty, ovr_slots, ovr_buckets, max_len,
-         100 * h->count / h->size,
-         h->stats.inserts, h->stats.searches,
-         h->stats.deletions, h->stats.rehashes);
-   }
-
-   str_append(buf, html_footer);
-   return (buf);
-}
-
-static struct str *
-html_tcp_stats(const char *ip)
-{
-   struct bucket *h = host_search(ip);
-
-   if (h == NULL)
-      return (NULL); /* no such host */
-   return html_hashtable_stats(h->u.host.ports_tcp);
-}
-
-static struct str *
-html_udp_stats(const char *ip)
-{
-   struct bucket *h = host_search(ip);
-
-   if (h == NULL)
-      return (NULL); /* no such host */
-   return html_hashtable_stats(h->u.host.ports_udp);
-}
-
-static struct str *
-html_proto_stats(const char *ip)
-{
-   struct bucket *h = host_search(ip);
-
-   if (h == NULL)
-      return (NULL); /* no such host */
-   return html_hashtable_stats(h->u.host.ip_protos);
-}
-
 /* ---------------------------------------------------------------------------
  * Database import and export code:
- * Initially written and contributed by Ben stewart.
+ * Initially written and contributed by Ben Stewart.
  * copyright (c) 2007 Ben Stewart, Emil Mikulic.
  */
 static int hosts_db_export_ip(const struct hashtable *h, const int fd);
 static int hosts_db_export_tcp(const struct hashtable *h, const int fd);
 static int hosts_db_export_udp(const struct hashtable *h, const int fd);
 
-static const char export_proto_ip  = 'P';
-static const char export_proto_tcp = 'T';
-static const char export_proto_udp = 'U';
+static const char
+   export_proto_ip  = 'P',
+   export_proto_tcp = 'T',
+   export_proto_udp = 'U';
 
-static const unsigned char export_tag_host_ver1[] = {'H', 'S', 'T', 0x01};
+static const unsigned char
+   export_tag_host_ver1[] = {'H', 'S', 'T', 0x01},
+   export_tag_host_ver2[] = {'H', 'S', 'T', 0x02};
 
 /* ---------------------------------------------------------------------------
  * Load a host's ip_proto table from a file.
@@ -1226,15 +1155,31 @@ hosts_db_import_host(const int fd)
    uint8_t hostname_len;
    uint64_t in, out;
    unsigned int pos = xtell(fd);
+   char hdr[4];
+   int ver = 0;
 
-   if (!read_file_header(fd, export_tag_host_ver1)) return 0;
+   if (!readn(fd, hdr, sizeof(hdr))) return 0;
+   if (memcmp(hdr, export_tag_host_ver2, sizeof(hdr)) == 0)
+      ver = 2;
+   else if (memcmp(hdr, export_tag_host_ver1, sizeof(hdr)) == 0)
+      ver = 1;
+   else {
+      warnx("bad host header: %02x%02x%02x%02x",
+         hdr[0], hdr[1], hdr[2], hdr[3]);
+      return 0;
+   }
 
-   /* HOST ENTRY */
    if (!readaddr(fd, &addr)) return 0;
    verbosef("at file pos %u, importing host %s", pos, ip_to_str(addr));
    host = host_get(addr);
    assert(host->u.host.ip == addr); /* make fn? */
 
+   if (ver > 1) {
+      uint64_t t;
+      if (!read64(fd, &t)) return 0;
+      host->u.host.last_seen = (time_t)t;
+   }
+
    assert(sizeof(host->u.host.mac_addr) == 6);
    if (!readn(fd, host->u.host.mac_addr, sizeof(host->u.host.mac_addr)))
       return 0;
@@ -1302,14 +1247,13 @@ int hosts_db_export(const int fd)
    for (i = 0; i<hosts_db->size; i++)
    for (b = hosts_db->table[i]; b != NULL; b = b->next) {
       /* For each host: */
-
-      if (!writen(fd, export_tag_host_ver1, sizeof(export_tag_host_ver1)))
+      if (!writen(fd, export_tag_host_ver2, sizeof(export_tag_host_ver2)))
          return 0;
 
-      /* HOST ENTRY */
       if (!writeaddr(fd, b->u.host.ip)) return 0;
 
-      /* MACADDR */
+      if (!write64(fd, (uint64_t)(b->u.host.last_seen))) return 0;
+
       assert(sizeof(b->u.host.mac_addr) == 6);
       if (!writen(fd, b->u.host.mac_addr, sizeof(b->u.host.mac_addr)))
          return 0;
