@@ -1,5 +1,5 @@
 /* darkstat 3
- * copyright (c) 2001-2007 Emil Mikulic.
+ * copyright (c) 2001-2008 Emil Mikulic.
  *
  * http.c: embedded webserver.
  * This borrows a lot of code from darkhttpd.
@@ -36,9 +36,9 @@ static const char mime_type_xml[] = "text/xml";
 static const char mime_type_html[] = "text/html; charset=us-ascii";
 static const char mime_type_css[] = "text/css";
 static const char mime_type_js[] = "text/javascript";
-static const char encoding_deflate[] =
+static const char encoding_gzip[] =
     "Vary: Accept-Encoding\r\n"
-    "Content-Encoding: deflate\r\n";
+    "Content-Encoding: gzip\r\n";
 
 static const char server[] = PACKAGE_NAME "/" PACKAGE_VERSION;
 static int idletime = 60;
@@ -66,7 +66,7 @@ struct connection {
     /* char request[request_length+1] is null-terminated */
     char *request;
     size_t request_length;
-    int accept_deflate;
+    int accept_gzip;
 
     /* request fields */
     char *method, *uri, *query; /* query can be NULL */
@@ -261,7 +261,7 @@ static struct connection *new_connection(void)
     conn->last_active = now;
     conn->request = NULL;
     conn->request_length = 0;
-    conn->accept_deflate = 0;
+    conn->accept_gzip = 0;
     conn->method = NULL;
     conn->uri = NULL;
     conn->query = NULL;
@@ -435,7 +435,7 @@ static char *parse_field(const struct connection *conn, const char *field)
 /* ---------------------------------------------------------------------------
  * Parse an HTTP request like "GET /hosts/?sort=in HTTP/1.1" to get the method
  * (GET), the uri (/hosts/), the query (sort=in) and whether the UA will
- * accept deflate encoding.  Remember to deallocate all these buffers.  Query
+ * accept gzip encoding.  Remember to deallocate all these buffers.  Query
  * can be NULL.  The method will be returned in uppercase.
  */
 static int parse_request(struct connection *conn)
@@ -478,8 +478,8 @@ static int parse_request(struct connection *conn)
     /* parse important fields */
     accept_enc = parse_field(conn, "Accept-Encoding: ");
     if (accept_enc != NULL) {
-        if (strstr(accept_enc, "deflate") != NULL)
-            conn->accept_deflate = 1;
+        if (strstr(accept_enc, "gzip") != NULL)
+            conn->accept_gzip = 1;
         free(accept_enc);
     }
     return (1);
@@ -516,17 +516,17 @@ static_graph_js(struct connection *conn)
 }
 
 /* ---------------------------------------------------------------------------
- * Deflate a reply, if requested and possible.  Don't bother with a minimum
- * length requirement since I've never seen a page fail to compress.
+ * gzip a reply, if requested and possible.  Don't bother with a minimum
+ * length requirement, I've never seen a page fail to compress.
  */
 static void
-process_deflate(struct connection *conn)
+process_gzip(struct connection *conn)
 {
     char *buf;
     size_t len;
     z_stream zs;
 
-    if (!conn->accept_deflate)
+    if (!conn->accept_gzip)
         return;
 
     buf = xmalloc(conn->reply_length);
@@ -536,7 +536,10 @@ process_deflate(struct connection *conn)
     zs.zfree = Z_NULL;
     zs.opaque = Z_NULL;
 
-    if (deflateInit(&zs, Z_BEST_COMPRESSION) != Z_OK)
+    if (deflateInit2(&zs, Z_BEST_COMPRESSION, Z_DEFLATED,
+        15+16, /* 15 = biggest window, 16 = add gzip header+trailer */
+        8 /* default */,
+        Z_DEFAULT_STRATEGY) != Z_OK)
        return;
 
     zs.avail_in = conn->reply_length;
@@ -559,17 +562,8 @@ process_deflate(struct connection *conn)
 
     conn->reply = buf;
     conn->reply_length -= zs.avail_out;
-    conn->encoding = encoding_deflate;
+    conn->encoding = encoding_gzip;
     deflateEnd(&zs);
-    verbosef("compressed %u bytes to %u bytes",
-       (unsigned int)len, (unsigned int)conn->reply_length);
-
-    /* To unconfuse IE (and apparently Safari), we must trim 2 bytes from
-     * the header.  It's also been suggested to trim 5 from the footer
-     * but I've found that this breaks things.
-     */
-    conn->reply_sent = 2;
-    conn->reply_length -= 2;
 }
 
 /* ---------------------------------------------------------------------------
@@ -580,8 +574,8 @@ static void process_get(struct connection *conn)
     char *decoded_url, *safe_url;
     char date[DATE_LEN];
 
-    verbosef("http: [%s %s [%s]]", conn->method, conn->uri,
-        (conn->query == NULL)?"<null>":conn->query);
+    verbosef("http: %s \"%s\" %s", conn->method, conn->uri,
+        (conn->query == NULL)?"":conn->query);
 
     /* work out path of file being requested */
     decoded_url = urldecode(conn->uri);
@@ -630,7 +624,7 @@ static void process_get(struct connection *conn)
     }
     free(safe_url);
 
-    process_deflate(conn);
+    process_gzip(conn);
     assert(conn->mime_type != NULL);
     conn->header_length = xasprintf(&(conn->header),
         "HTTP/1.1 200 OK\r\n"
@@ -744,10 +738,7 @@ static void poll_send_header_and_reply(struct connection *conn)
     assert(conn->reply_length > 0);
     assert(conn->header_sent == 0);
 
-    assert((conn->reply_sent == 0) || (conn->reply_sent == 2));
-    /* reply_sent can be 2 if we're trying to chop off the first two bytes of
-     * a deflated reply.  Blame Internet Explorer!!!
-     */
+    assert(conn->reply_sent == 0);
 
     /* Fill out iovec */
     iov[0].iov_base = conn->header;
@@ -778,7 +769,6 @@ static void poll_send_header_and_reply(struct connection *conn)
         return;
     }
     /* else */
-    verbosef("completely sent header");
     conn->header_sent = conn->header_length;
     sent -= conn->header_length;
 
@@ -789,7 +779,6 @@ static void poll_send_header_and_reply(struct connection *conn)
         return;
     }
     /* else */
-    verbosef("completely sent reply");
     conn->reply_sent = conn->reply_length;
     conn->state = DONE;
 }
